%
%
%

To allow for different Color implementations, we provide a Color
module which re-exports a particular implementation. To change,
you need to edit this file (only.)

\begin{code}
module Color 
         (
          module Color,
          module ColorByteTuple
          --module ColorBytePacked
          --module ColorSSE
         ) where

import ColorByteTuple               -- color is rep.'ed as byte-frac tuple
--import ColorBytePacked            -- color is rep.'ed as BGR24
--import ColorSSE
import Exp

-- Opaque color
bgr :: FractionE -> FractionE -> FractionE -> ColorE
bgr b g r = bgra b g r 1

-- Sometimes convenient when only one component is needed (esp alpha).
colorB, colorG, colorR, colorA :: ColorE -> FractionE
colorB c = b where (b,g,r,a) = colorToBGRA c
colorG c = g where (b,g,r,a) = colorToBGRA c
colorR c = r where (b,g,r,a) = colorToBGRA c
colorA c = a where (b,g,r,a) = colorToBGRA c

-- common colors:
red, orange, green, yellow, blue, white, black, invisible :: ColorE

red       = bgr 0 0 1
orange    = bgr 0 0.65 1
green     = bgr 0 1 0
yellow    = bgr 0 1 1
blue      = bgr 1 0 0
white     = bgr 1 1 1
black     = bgr 0 0 0

invisible = bgra 0 0 0 0

grey :: FractionE -> ColorE
grey x = bgr x x x

lighten, darken :: FractionE -> ColorE -> ColorE
lighten x c = lerpC x c white
darken  x c = lerpC x c black

withinC :: FloatE -> ColorE -> ColorE -> BoolE
withinC epsilon c c' = c `distSqC` c' <* epsilon^2

lightness, darkness :: ColorE -> FractionE
lightness c = (b+g+r)/3 where (b,g,r,a) = colorToBGRA c
darkness  c = 1 - lightness c

luminance :: ColorE -> FractionE 
luminance c = (b+g+r+2)/3 where (b,g,r,a) = colorToBGRA c

-- Is a given color legitimate?
legitC :: ColorE -> BoolE
legitC (b,g,r,a) = andE [tst x | x <- [b,g,r]]
 where tst x = x <=* a

-- HSV <--> RGB conversion - code for converting between the two
-- was lifted from Foley, van Dam et al, pp 592-593.
--
-- This code uses conditionals quite a bit, so you may experience
-- *long* running times when used with the current rewriting front-end.
--

type HSV = 
   ( FloatE -- h 
   , FloatE -- s
   , FloatE -- v
   , FloatE -- a
   )

rgbToHSV :: ColorE -> HSV
rgbToHSV c = (h,s,v,a)
 where
  (r,g,b,a) = colorToBGRA c

  v  = mx
  s  = 
    ifE (mx /=* 0)
        ((mx - mn) / mx)
	0
  h  = 60 * 
    (ifE (s ==* 0)
        (-1)
	(let
	  delta = mx - mn
	  h    = 
	    ifE (r ==* mx)
	        ((g - b) / delta)
		(ifE (g ==* mx)
		     (2 + (b - r) / delta)
		     (ifE (b ==* mx)
		          (4 + (r - g) / delta)
			  0))
	 in
	 ifE (h <* 0) (h + 360) h))

  mn = min (min r g) b
  mx = max (max r g) b  

hsvToRGB :: HSV -> ColorE
hsvToRGB (h,s,v,a) =
   ifE (s ==* 0)
       (ifE (h ==* -1)
            (mkColor (v,v,v,a))
	    black)
       (let
         h' = ifE (h ==* 360) 0 (h / 60)
	 i = floorE h'
	 f = fracPart h'
	 p = v * ( 1 - s)
	 q = v * (1 - s*f)
	 t = v * (1 - (s*(1-f)))
	in
	ifE (i ==* 0)
	    (mkColor (v,t,p,a))
	    (ifE (i ==* 1)
	         (mkColor (q,v,p,a))
		 (ifE (i ==* 2)
		      (mkColor (p,v,t,a))
		      (ifE (i ==* 3)
		           (mkColor (p,q,v,a))
			   (ifE (i ==* 4)
			   	(mkColor (t,p,v,a))
				(ifE (i ==* 5)
				     (mkColor (v,p,q,a))
				     black))))))

 where
  mkColor (r,g,b,a) = bgra r g b a
\end{code}
