{-
  misc. neighborhood filters, including a rank-value / median filter.
-}
module Neighbor where

import Exp
import Color
import Image
import Arrays
import Bitmap
import UI

{-
  Neighborhood averaging filter - kernel/mask size can
  be adjusted via a UI control.
-}
avgUI :: UI ImageC
avgUI = do 
  v  <- islider "mask size" (3,25) 3
  im <- imageSourceD "input image"
  return (avgFilter 1 v im)

{-
  3x3 softening filter example.
-}
softenUI :: UI ImageC
softenUI = do
  g  <- slider "gain factor" (-1,10) 1
  im <- imageSourceD "input image"
  return (softenFilter g im)

{-
 3x3 Laplacian filter example
-}
laplacianUI :: UI ImageC
laplacianUI = do
  im <- imageSourceD "input image"
  return (laplacian im)

{-
 3x3 Line detector:
-}
lineDetectUI :: UI ImageC
lineDetectUI = do
  kind <- radioBoxInt "Direction" ["Horiz", "Vert", "Left", "Right"] 0
  im   <- imageSourceD "input image"
  return (lineDetect kind im)

{-
 High boost filter example.
-}
boostUI :: UI ImageC
boostUI = do
  g  <- slider "gain factor" (0,10) 1
  im <- imageSourceD "input image"
  return (boostFilter g im)

boostTimeUI :: UI ImageBC
boostTimeUI = do
  im <- imageSourceD "input image"
  return (\ t -> boostFilter (0.6 + 0.5*sin t) im)

{-
 3x3 sharpening filter.
-}
sharpUI :: UI ImageC
sharpUI = do
  g  <- slider "gain factor" (0,10) 1
  im <- imageSourceD "input image"
  return (sharpenFilter g im)

{-
 Rank value filter, i.e., given the 8 pixels surrounding a pixel, order
 them by lightness & pick a member of that ordering as the result pixel
 value.

 So, by adjusting the UI control you'll get the min, max and median filters.
-}
rankValueUI :: UI ImageC
rankValueUI = do
  r  <- islider "rank value" (0,8) 4
  im <- imageSourceD "input image"
  return (rankValue 3 r im)
   
-- 3x3 softening filter.
softenFilter :: FloatE -> Array2 ColorE -> ImageC
softenFilter g = neighbor (4+g) toBGRA32 id (reifyArray2 (Array2 3 3 f))
 where
   f (x,y) = 
     ifE (x ==* 1 &&* y ==* 1)
     	 g
	 (ifE (x ==* 1 ||* y ==* 1)
	      1
	      0)

-- 3x3 Laplacian (2nd derivative - enhances changes)
laplacian :: Array2 ColorE -> ImageC
laplacian = neighbor 1 lightness adjust (reifyArray2 (Array2 3 3 f))
  where
   adjust v = toBGRA32 (grey (1 - max 0 (min 1 v)))

   f (x,y) =
     ifE (x ==* 1 &&* y ==* 1)
     	 4
	 (ifE (x ==* 1 ||* y ==* 1)
	      (-1)
	      0)

-- 3x3 line detector
lineDetect :: IntE -> Array2 ColorE -> ImageC
lineDetect kind = neighbor 1 lightness adjust arr
  where
   adjust v = toBGRA32 (grey (1 - max 0 (min 1 v)))

   arr =
       reifyArray2 (Array2 3 3 $
 	    ifE (kind ==* 0)
	        fHoriz
		(ifE (kind ==* 1)
		    fVert
		    (ifE (kind ==* 2)
		         fLeft
		         fRight)))
		    
   fHoriz (x,y) = ifE (x ==* 1) 2 (-1)
   fVert (x,y)  = ifE (y ==* 1) 2 (-1)

   fLeft (x,y)  = ifE (x ==* y) 2 (-1)
   fRight (x,y) = ifE ((x+y) ==* 2) 2 (-1)


-- 3x3 high-boost filter.
boostFilter :: FloatE -> Array2 ColorE -> ImageC
boostFilter g = neighbor 9 lightness adjust (reifyArray2 (Array2 3 3 f))
 where
   adjust v = toBGRA32 (grey ((v+1) / g))

   f (x,y) = 
     ifE (x ==* 2 &&* y ==* 2)
     	 (9*g-1)
	 (-1)

-- 3x3 sharpening filter.
sharpenFilter :: FloatE -> Array2 ColorE -> ImageC
sharpenFilter g = neighbor 1 toBGRA32 scaleRes (reifyArray2 (Array2 3 3 f))
 where
   scaleRes v = f2i ((i2f v + 4*g) / 5*g)

   f (x,y) = 
     ifE (x ==* 1 &&* y ==* 1)
     	 (1+4*g)
	 (ifE (x ==* 1 ||* y ==* 1)
	      (-g)
	      0)


-- n'borhood averaging filter.
avgFilter :: FloatE -> IntE -> Array2 ColorE -> ImageC
avgFilter val v = neighbor (i2f v * i2f v * val) toBGRA32 id (reifyArray2 (Array2 v v f))
 where
   f (x,y) = val

avgFilterD :: FloatE -> IntE -> Array2 ColorE -> Array2 ColorE
avgFilterD val sz arr = arr2Map fromBGRA32 (neighborD (i2f sz * i2f sz * val) toBGRA32 (Array2 sz sz f) arr)
 where
   f (x,y) = val

-- rank-value filter.
rankValue :: IntE -> IntE -> Array2 ColorE -> ImageC
rankValue sz rank arr = reconstruct (arr2Map fromBGRA32 medianArr)
 where
   medianArr = reifyArray2 (rankValueD sz rank arr)

-- using a sz*sz kernel, rank the pixel values & pick one of them.
rankValueD :: IntE -> IntE -> Array2 ColorE -> Array2 BGRA32E
rankValueD sz rank (Array2 iw ih img) = Array2 iw' ih' img'
 where
   iw' = iw+sz
   ih' = ih+sz
   r   = sz `div` 2
   arr = performAction (newArr (sz*sz) 0)
   img' (x,y) = 
    performAction $
      (loopLHInc (-r) r $ \ dy ->
       loopLHInc (-r) r $ \ dx ->      
         writeArr arr ((dy+r)*sz + (dx+r)) (toBGRA32 (imgSafe (x+dx,y+dy)))) >>*
	-- rank the pixel values
       sortArray (\ x -> lightness (fromBGRA32 x)) arr (sz*sz) >>=* \ arr -> 
       readMutArr arr rank

   inBounds (i, j) = 0 <=* i &&* i <* iw &&* 0 <=* j &&* j <* ih
   imgSafe = cond inBounds img empty

-- selection sort
--sortArray :: (BGRA32E -> FloatE) -> ArrE a -> IntE -> Exp (Action (ArrE a))
sortArray f arr sz = 
   newRef 0  >>=* \ ref ->
   (loop sz $ \ i ->
     writeRef ref i >>*
     (loopLHInc (i+1) sz $ \ j ->
        readRef ref >>=* \ x ->
	readMutArr arr j >>=* \ j_val ->
	readMutArr arr x >>=* \ x_val ->
        ifE (f j_val <* f x_val)
    	    (writeRef ref j)
	    noop) >>*
     readRef ref >>=* \ x -> 
     readMutArr arr x >>=* \ x_val ->
     readMutArr arr i >>=* \ i_val ->
     writeArr arr x i_val >>*
     writeArr arr i x_val) >>*
  returnE arr

-- spatial neighborhood filters, generalised version of
-- Convolve.convolveDC
neighbor :: (Num (Exp a), ExpType a)
	 => FloatE 
	 -> (ColorE -> Exp a)
	 -> (Exp a -> BGRA32E)
	 -> Array2 FloatE 
	 -> Array2 ColorE 
	 -> ImageC
neighbor sumCoeff colorMap f mask imD = reconstruct (arr2Map (fromBGRA32.f) neighbored)
 where
  neighbored = reifyArray2 (neighborD sumCoeff colorMap mask imD)

neighborD :: (Num (Exp a), ExpType a )
	  => FloatE 
	  -> (ColorE -> Exp a)
          -> Array2 FloatE 
	  -> Array2 ColorE 
	  -> Array2 (Exp a)
neighborD sumCoeff colorMap (Array2 mw mh mask) (Array2 iw ih img) = Array2 iw' ih' img'
 where
   iw' = iw+mw
   ih' = ih+mh
   rw  = mw `div` 2
   rh  = mh `div` 2
   img' (x,y) = 
     sumLoopLHInc (-rh) rh $ \ dy ->
     sumLoopLHInc (-rw) rw $ \ dx ->
     	 colorMap ((invTot * mask (rw+dx,rw+dy)) *^ imgSafe (x+dx,y+dy))

   invTot = recip sumCoeff
   inBounds (i, j) = 0 <=* i &&* i <* iw &&* 0 <=* j &&* j <* ih
   imgSafe = cond inBounds img empty

(*^) :: FloatE -> ColorE -> ColorE
c *^ (b,g,r,a) = --trace ("*^ " ++ show c) $
                 (f b, f g, f r, f a)
 where
   f = (toBF c `multBF`)
